;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "easy-hugo" "3.13.64"
  "Write blogs made with hugo by markdown or org-mode."
  '((emacs     "25.1")
    (request   "0.3.0")
    (transient "0.3.6"))
  :url "https://github.com/masasam/emacs-easy-hugo"
  :commit "f39adb614ce2c8969aeb7198e9677a189e391a14"
  :revdesc "f39adb614ce2")
